package de.hawhamburg.reachability.reasoner;

import java.util.Random;

import com.google.gson.Gson;

import de.hawhamburg.livingplace.messaging.activemq.wrapper.LPPublisher;
import de.hawhamburg.livingplace.messaging.activemq.wrapper.LPSubscriber;
import de.hawhamburg.reachability.ProjectInfo;
import de.hawhamburg.reachability.StopableObject;
import de.hawhamburg.reachability.conf.MainConfig;
import de.hawhamburg.reachability.conf.ReasonerConfig;
import de.hawhamburg.reachability.conf.enumeration.ConfigType;
import de.hawhamburg.reachability.json.JsonReachabilityInformation;
import de.hawhamburg.reachability.json.JsonReachabilityInstruction;
import de.hawhamburg.reachability.json.JsonReachabilitySensorData;
import de.hawhamburg.reachability.properties.PropertyService;
import de.hawhamburg.reachability.properties.PropertyServiceNotifiable;
import de.hawhamburg.reachability.registry.enumeration.RegistryAgentType;
import de.hawhamburg.reachability.registry.enumeration.RegistryMessageType;
import de.hawhamburg.reachability.registry.json.JsonRARegistryMessage;
import de.hawhamburg.reachability.util.ConnectionUtil;
import de.hawhamburg.reachability.util.OutputUtil;

/**
 * Abstract class which needs to be extended by each Reasoner.<br><br>
 * 
 * This abstract class needs to be extended by a Reasoner to be used in common with the reachability project. It 
 * provides some ActiveMQ connection relevant methods, to extremely simplify the sensor adapter design.
 * 
 * @author  {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_AUTHOR} 
 * @version {@value de.hawhamburg.reachability.ProjectInfo#REACHABILITY_AGENT_VERSION}
 */
public abstract class Reasoner extends    Thread
                               implements StopableObject,
                                          ReasoningEngine,
                                          PropertyServiceNotifiable {

    /**
     * Holds the activity state of this object. If silent is <code>true</code>, the worker won't publish its results on 
     * the result topic.
     */
    private boolean silent                          = true;
    
    /**
     * The name of the ActiveMQ topic on which user feedback will be propagated. 
     */
    private String feedbackTopicName                = null;
    
    /**
     * The name of the ActiveMQ topic on which the sensor delivers its measuring results. 
     */
    private String resultTopicName                  = null;
    
    /**
     * The name of the ActiveMQ topic on which the sensor publishes its generalized data set.
     */
    private String internalSensorDataTopicName      = null;
    
    /**
     * The name of the ActiveMQ topic on which the systems registry service invites other components to join him.
     */
    private String registryRequestTopicName         = null;
    
    /**
     * The name of the ActiveMQ topic on which this {@link Reasoner} sends its register messages to the systems registry
     * service.
     */
    private String registryTopicName                = null;
    
    /**
     * The {@link LPSubscriber}, which will get user feedback from user interfaces.
     */
    private LPSubscriber feedbackTopic              = null;
    
    /**
     * The {@link LPPublisher}, which will write the results to the agents result.
     */
    private LPPublisher resultTopic                 = null;
    
    /**
     * The {@link LPPublisher}, which will get the generalized result set from {@link #internalSensorDataTopicName}.
     */
    private LPSubscriber internalSensorDataTopic    = null;
    
    /**
     * The {@link LPSubscriber}, which listens for register requests from the systems registry service via the ActiveMQ.
     */
    private LPSubscriber requestRegistryTopic       = null;
    
    /**
     * The {@link LPPublisher}, which sends register messages to the systems registry service via the ActiveMQ.
     */
    private LPPublisher registryTopic               = null;

    /**
     * Determines if the method {@link #shutDown()} has been called.
     */
    private boolean shutDown                        = false;
    
    /**
     * Determines if the {@link Reasoner} has stopped.
     */
    private boolean stopped                         = false;
    
    /**
     * Local random number generator.
     */
    private Random rand                             = null;
    
    /**
     * The {@link PropertyService} of this {@link Reasoner}, which manages property requests and changes from user 
     * interfaces.
     */
    private PropertyService propertyService         = null;
    
    /**
     * Protected constructor ensures, that there will be no direct instance of this class.
     */
    protected Reasoner() {
        this.rand                           = new Random();
        this.feedbackTopicName              = MainConfig.get("REACHABILITY_AGENT_TOPIC");
        this.resultTopicName                = MainConfig.get("REACHABILITY_AGENT_RESULT_TOPIC");
        this.internalSensorDataTopicName    = MainConfig.get("INTERNAL_SENSOR_DATA_TOPIC");
        this.registryRequestTopicName       = MainConfig.get("REACHABILITY_REGISTRY_REQUEST_TOPIC");
        this.registryTopicName              = MainConfig.get("REACHABILITY_REGISTRY_TOPIC");
    }
    
    /**
     * Set the silent state. If silent is set to <code>true</code>, the worker won't publish his informations on the
     * general result topic.
     * 
     * @param state The current state of silence as {@link Boolean}.
     */
    public final void setSilent(final boolean state) {
        this.silent = state;
    }
    
    /**
     * Checks if the current workers running mode is silent or not.
     * 
     * @return      The current state of silence as {@link Boolean}.
     */
    public final boolean isSilent() {
        return this.silent;
    }

    /**
     * @see de.hawhamburg.reachability.StopableObject#isActive()
     * 
     * @return      <code>true</code> if {@link #shutDown()} wasn't called, otherwise <code>false</code>.
     */
    public final boolean isActive() {
        return !this.shutDown;
    }
    
    /**
     * Determines if the {@link Reasoner} has stopped.
     * 
     * @return      <code>true</code> if {@link Reasoner} has stopped, otherwise <code>false</code>.
     */
    public final boolean stopped() {
        return this.stopped;
    }

    /**
     * @see de.hawhamburg.reachability.StopableObject#shutDown()
     */
    public final void shutDown() {
        this.shutDown = true;
    }
    
    /**
     * Frequently check for sensor data updates. On success notify the reasoner.
     */
    public final void run() {
        // Initialize connection
        init();
        
        // Send initial registration message to reachability agent
        sendInitialRegisterMessage();
        
        // Life time...
        while (this.isActive()) {
            // Look for register request
            handleRegisterRequestMessage();
            
            // Handle active user feedback
            handleActiveUserFeedback();
            
            // Get sensor data and publish them into knowledge base
            handleInternalSensorData();
            
            // Publish reachability states
            if (!this.isSilent()) {
                publishReachabilityStatus();
            }
        }
        end();
    }
    
    /**
     * Initialize all required ActiveMQ connections and starts the property service of this {@link Reasoner}.
     */
    private void init() {
        this.feedbackTopic              = ConnectionUtil.getSubscriber(feedbackTopicName, this);
        this.resultTopic                = ConnectionUtil.getPublisher(this.resultTopicName, this);
        this.internalSensorDataTopic    = ConnectionUtil.getSubscriber(this.internalSensorDataTopicName, this);
        this.requestRegistryTopic       = ConnectionUtil.getSubscriber(this.registryRequestTopicName, this);
        this.registryTopic              = ConnectionUtil.getPublisher(this.registryTopicName, this);
        
        // Start property service
        String descPath = ReasonerConfig.get("REASONING_ENGINE_PROPERTY_DESCRIPTION_SOURCE");
        if (descPath != null) {
            this.propertyService = new PropertyService(this, ConfigType.REASONER, descPath);
        } else {
            this.propertyService = new PropertyService(this, ConfigType.REASONER);
        }
        this.propertyService.start();
    }
    
    /**
     * Sends an initial register message by using {@link #sendRegistryMessage(RegistryMessageType)} with 
     * {@link RegistryMessageType#REGISTER}.
     */
    private void sendInitialRegisterMessage() {
        sendRegistryMessage(RegistryMessageType.REGISTER);
    }
    
    /**
     * Checks for new registry invitations from the systems registry service and sends register message if there's 
     * some by using {@link #sendRegistryMessage(RegistryMessageType)} with {@link RegistryMessageType#REGISTER}.
     */
    private void handleRegisterRequestMessage() {
        String json = ConnectionUtil.getMessageOnTopic(requestRegistryTopic);
        if (json != null) {
            OutputUtil.info("  Recived register request from registry...");
            sendRegistryMessage(RegistryMessageType.REGISTER);
        }
    }
    
    /**
     * Handles active user feedback, if some new arises via the ActiveMQ.
     */
    private void handleActiveUserFeedback() {
        String msg = ConnectionUtil.getMessageOnTopic(feedbackTopic);
        if (msg != null) {
            Gson g = new Gson();
            JsonReachabilityInstruction feedback = g.fromJson(msg, JsonReachabilityInstruction.class);
            this.handleFeedback(feedback.Group, feedback.Status);
        }
    }
    
    /**
     * Forward internal sensor data to the reasoning engine, if some new arises via the ActiveMQ.
     */
    private void handleInternalSensorData() {
        String msg = ConnectionUtil.getMessageOnTopic(internalSensorDataTopic);
        if (msg != null) {
            Gson g = new Gson();
            JsonReachabilitySensorData sensorData = g.fromJson(msg, JsonReachabilitySensorData.class);
            this.setSensorData(sensorData);
        }
    }
    
    /**
     * Extracts dimension and sensor data from a received {@link JsonReachabilitySensorData} and adds them to the 
     * working memory.
     * 
     * @param data  The {@link JsonReachabilitySensorData} which has been measured.
     */
    protected abstract void setSensorData(JsonReachabilitySensorData data);

    /**
     * The method will publish the current reachability status using the publisher {@link #resultTopic} and
     * {@link #getJsonReachabilityInformation()}.
     */
    private void publishReachabilityStatus() {
        JsonReachabilityInformation information = getJsonReachabilityInformation();
        ConnectionUtil.setMessageOnTopic(this.resultTopic, information);
    }

    /**
     * Returns a single {@link JsonReachabilityInformation} object which contains the 
     * {@link de.hawhamburg.reachability.enumeration.ReachabilityStatus} to all
     * {@link de.hawhamburg.reachability.enumeration.ReachabilityGroup}s and different identification informations.
     * (For further informations see {@link JsonReachabilityInformation}).
     *
     * @return      A new JSON object which can be published on an ActiveMQ as {@link JsonReachabilityInformation}.
     */
    private JsonReachabilityInformation getJsonReachabilityInformation() {
        JsonReachabilityInformation msg = new JsonReachabilityInformation();

        msg.Id              = this.getID();
        msg.Version         = this.getVersion();
        msg.currentStatus   = this.getEstimatedReachability();
        msg.Remark          = "just random testing stuff";

        return msg;
    }
    
    /**
     * Unregisters from systems registry service, disconnects all ActiveMQ topic connections and sets {@link #stopped}
     * to <code>true</code>.
     */
    private void end() {
        // Unregister
        sendRegistryMessage(RegistryMessageType.UNREGISTER);
        
        // Free system resources
        this.cleanUp();

        // End property service
        this.propertyService.shutDown();
        
        // Disconnection AMQ connections
        if (feedbackTopic != null) {
            feedbackTopic.disconnect();
        }
        
        if (resultTopic != null) {
            resultTopic.disconnect();
        }
        
        if (internalSensorDataTopic != null) {
            internalSensorDataTopic.disconnect();
        }
        
        if (requestRegistryTopic != null) {
            requestRegistryTopic.disconnect();
        }
        
        if (registryTopic != null) {
            registryTopic.disconnect();
        }

        this.stopped = true;
    }

    /**
     * Cleans up by freeing all resources.
     */
    protected abstract void cleanUp();
    
    /**
     * Creates and sends a new {@link JsonRARegistryMessage} onto the {@link #registryTopic}. This may register or
     * unregister this {@link Reasoner} due to the supplied {@link RegistryMessageType}.
     * 
     * @param type  The {@link RegistryMessageType} which indicates if the {@link Reasoner} shall be registered or 
     *              unregistered from the systems registry service.
     */
    private void sendRegistryMessage(final RegistryMessageType type) {
        JsonRARegistryMessage register = new JsonRARegistryMessage();
        
        register.Id             = this.getID();
        register.Version        = this.getVersion();
        register.MessageType    = type;
        register.AgentType      = RegistryAgentType.REASONER;
        register.Name           = this.getClass().getSimpleName();
        register.Remark         = "Here am I...";
        
        ConnectionUtil.setMessageOnTopic(registryTopic, register);

        if (type.equals(RegistryMessageType.REGISTER)) {
            OutputUtil.info("+ Sending new REASONER registration: '" + register.Name + "'");
        } else {
            OutputUtil.info("- Sending REASONER unregistration: '" + register.Name + "'");
        }
    }
    
    /**
     * Push current properties to other components via the ActiveMQ.
     */
    protected final void pushProperties() {
        this.propertyService.pushProperties(null);
    }
    
    /**
     * Returns the ID of this {@link Reasoner}.
     * 
     * @return      The ID of this {@link Reasoner} as {@link String}.
     */
    protected final String getID() {
        return ProjectInfo.REACHABILITY_AGENT_NAME + "_Reasoner_" + this.getClass().getSimpleName() + "_"
                + this.rand.nextInt(Integer.MAX_VALUE);
    }
    
    /**
     * Returns the version of this {@link Reasoner}.
     * 
     * @return      The version of this {@link Reasoner} as {@link String}.
     */
    protected final String getVersion() {
        return ProjectInfo.REACHABILITY_AGENT_VERSION;
    }
}
